/*
 *	On-screen Display
*
 *	(c) 2013--2014 Martin Mares <mj@ucw.cz>
 *	(c) 2020--2021 Jiri Kalvoda <jirikalvoda@kam.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "osdd-set.h"
#include "util.h"

int osd_set_trie_char_to_index(char in)
{
	if('0'<=in && in<='9') return in-'0';
	if(in == '_') return 10;
	if('a'<=in && in<='z') return in-'a'+11;
	return -1;
}

struct osd_set_trie * osd_set_trie_new(void)
{
	struct osd_set_trie * r = xmalloc(sizeof(r[0]));
	memset(r, 0, sizeof(r[0]));
	return r;
}

struct osd_set_trie * osd_set_trie_find(struct osd_set_trie * root, char * str, bool create)
{
	while(*str)
	{
		int i = osd_set_trie_char_to_index(*str);
		if(i<0)
		{
			if(create)
			{
				die("osd_set_trie char %c not supported\n",*str);
			}
			else return 0;
		}
		if(!root->next[i])
		{
			if(create)
				root->next[i] = osd_set_trie_new();
			else return 0;
		}
		root = root->next[i];
		str++;
	}
	return root;
}


// ***************************************************************************************************************

struct osd_line *msg_add_line(struct msg *msg, enum osd_line_type type)
{
  VADDSIZE(msg->lines, 1);
  struct osd_line *l = &VLAST(msg->lines);
  l->type = type;
  l->fg_color = "green";
  l->outline_color = "black";
  l->outline_width = 0;

  switch (l->type)
    {
    case OSD_TYPE_TEXT:
      l->u.text = 0;
      break;
    case OSD_TYPE_PERCENTAGE:
    case OSD_TYPE_SLIDER:
      l->u.percent = 0;
      break;
    default:
      die("osd_add_line: unknown type %d", type);
    }

  return l;
}

void osd_set_arrive(struct osd_set *set, struct msg *msg)
{
	for(int i=0;i<VSIZE(set->elements);i++)
	{
		if(msg->output_active[i] && !set->elements[i].disabled)
			if(set->elements[i].arrive)
				msg->do_not_que |= set->elements[i].arrive(set->elements[i].context,msg->lines,VSIZE(msg->lines));
	}
}

void osd_set_show(struct osd_set *set, struct msg *msg)
{
	for(int i=0;i<VSIZE(set->elements);i++)
	{
		msg->output_active[i] &= !set->elements[i].disabled;
		if(msg->output_active[i])
			if(set->elements[i].show) set->elements[i].show(set->elements[i].context,msg->lines,VSIZE(msg->lines));
	}
}

void osd_set_clear(struct osd_set *set, struct msg * msg)
{
	for(int i=0;i<VSIZE(set->elements);i++)
	{
		if(msg->output_active[i])
			if(set->elements[i].clear) set->elements[i].clear(set->elements[i].context);
	}
}

void osd_set_disable_outputs(struct osd_set *set, char * name, bool val)
{
	struct osd_set_trie *trie = osd_set_trie_find(&set->trie,name,0);
	if(trie)
	{
		for(int i=0;i<VSIZE(trie->vals);i++) set->elements[trie->vals[i]].disabled=val;
	}
}

void msg_active_outputs(struct osd_set *set, struct msg *msg, char * name, bool val)
{
	struct osd_set_trie *trie = osd_set_trie_find(&set->trie,name,0);
	if(trie)
	{
		for(int i=0;i<VSIZE(trie->vals);i++) msg->output_active[trie->vals[i]]=val;
	}
}

bool osd_set_handle_event(struct osd_set *set, XEvent *ev)
{
	bool r=0;
	for(int i=0;i<VSIZE(set->elements);i++)
	{
		if(set->elements[i].handle_event) r |= set->elements[i].handle_event(set->elements[i].context,ev);
	}
	return r;
}

struct osd_set osd_set_new(void)
{
  struct osd_set set;
  memset(&set, 0, sizeof(set));
  set.elements=0;
  return set;
}

void osd_set_add(struct osd_set *set, struct osd_abstract abs, int names_len, char ** names)
{
	for(int i=0;i<names_len;i++)
	{
		//fprintf(stderr,"NAME |%s|\n",names[i]);
		struct osd_set_trie *trie = osd_set_trie_find(&set->trie,names[i],1);
		VPB(trie->vals, VSIZE(set->elements));
	}
	VPB(set->elements, abs);
}

struct msg* osd_set_new_msg(struct osd_set * set)
{
	int size = sizeof(struct msg) + sizeof(bool) * VSIZE(set->elements);
	struct msg *msg = xmalloc(size);
	memset(msg, 0, size);
	VRESERVE(msg->lines, 4);
	return msg;
}
void msg_free(struct msg * msg)
{
	VFREE(msg->lines);
	if(msg->alocated_data) free(msg->alocated_data);
	free(msg);
}
