/*
 *	On-screen Display
 *
 *	(c) 2013--2014 Martin Mares <mj@ucw.cz>
 *	(c) 2020--2021 Jiri Kalvoda <jirikalvoda@kam.mff.cuni.cz>
 */

#ifndef OSDD_SET_H
#define OSDD_SET_H

#include <stdbool.h>
#include "util.h"

typedef struct _XDisplay Display;
typedef union _XEvent XEvent;
typedef struct _IO_FILE FILE;

struct display_state;

#define OSD_TRIE_LEN (26+10+1)

struct osd_set_trie
{
	struct osd_set_trie *next[OSD_TRIE_LEN];
	VECTOR(int) vals;
};

int osd_set_trie_char_to_index(char in);
struct osd_set_trie * osd_set_trie_new(void);
struct osd_set_trie * osd_set_trie_find(struct osd_set_trie * root, char * str, bool create);


enum osd_line_type {
  OSD_TYPE_TEXT,
  OSD_TYPE_PERCENTAGE,
  OSD_TYPE_SLIDER,
};

struct osd_line {
  enum osd_line_type type;
  char *fg_color;
  char *outline_color;
  int outline_width;
  union {					// Data dependent on type
    char * text;		// in UTF-8
    unsigned int percent;			// 0..100 for percentages and slider
  } u;
};

struct osd_abstract {
	void * context;

	bool disabled;
	void (*show)(void * context, struct osd_line * lines, int num_lines);
	bool (*arrive)(void * context, struct osd_line * lines, int num_lines);
	void (*clear)(void * context);
	bool (*handle_event)(void * context, XEvent *ev);
};

struct osd_set {
  VECTOR(struct osd_abstract) elements;
  struct osd_set_trie trie;
};

struct osd_creator_abstract {
	enum type {
		OSD_CEATE_ADD_ONE_OSD=0,
		OSD_CEATE_TYPE_ADD_MULTIPLE_OSD,
	} t;
	char *name;

	void (*help)(FILE * out);
	union {
	struct osd_abstract (*add_one_osd)(int argc, char ** argv,Display *dpy);
	void (*add_multiple_osd)(struct osd_set *set, int argc, char ** argv,Display *dpy,int names_len, char ** names);
	} new;
};

struct msg {
  int min_duration, max_duration; // ms
  timestamp_t start_light;
  VECTOR(struct osd_line) lines;

//internal usage only
  struct msg *next;
  VECTOR(char *) osdd_command_now;
  VECTOR(char *) osdd_command_after;

  void * alocated_data;
  bool do_not_que;
  bool output_active[];
};

struct osd_set osd_set_new(void);
void osd_set_add(struct osd_set *set, struct osd_abstract abs, int names_len, char ** names);
struct osd_line *msg_add_line(struct msg *set, enum osd_line_type type);
void osd_set_show(struct osd_set *set, struct msg *msg);
void osd_set_arrive(struct osd_set *set, struct msg *msg);
void osd_set_clear(struct osd_set *set, struct msg * msg);
bool osd_set_handle_event(struct osd_set *set, XEvent *ev);
void osd_set_disable_outputs(struct osd_set *set, char * name, bool val);
void msg_active_outputs(struct osd_set *set, struct msg *msg, char * name, bool val);
struct msg* osd_set_new_msg(struct osd_set * set);
void msg_free(struct msg * msg);


#define CREATE_STANDARD_CREATOR(NAME, SNAME) \
struct osd_creator_abstract NAME ## _creator_new(void) \
{ \
  struct osd_creator_abstract r; \
  memset(&r, 0, sizeof(r)); \
  r.name = SNAME; \
  r.help = NAME ## _new_help; \
  r.new.add_one_osd =  NAME ## _new; \
  return r; \
}

#define ADD_FUNC_TO_ABSTRACT(r, NAME) \
{ \
	r.context = state; \
	bool (*arrive)(struct NAME ## _state*, struct osd_line*, int) = NAME ## _arrive; \
	r.arrive = (bool (*)(void*, struct osd_line*, int)) arrive; \
	void (*show)(struct NAME ## _state*, struct osd_line*, int) = NAME ## _show; \
	r.show = (void (*)(void*, struct osd_line*, int)) show; \
	void (*clear)(struct NAME ## _state*) = NAME ## _clear; \
	r.clear = (void (*)(void*)) clear; \
}

#endif
