/*
 *	On-screen Display Daemon
 *
 *	(c) 2010--2014 Martin Mares <mj@ucw.cz>
 *	(c) 2020--2021 Jiri Kalvoda <jirikalvoda@kam.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <poll.h>
#include <getopt.h>
#include <locale.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <signal.h>
#include <sys/wait.h>

//#undef DEBUG
#include "util.h"
#include "osdd-set.h"

#include "display.h"
#include "osdd-file.h"
#include "osdd-mqtt.h"
#include "osdd-exec.h"
#include "osdd-system.h"


static struct osd_set osd;
Display * dpy;
#define FOREACHOSD for(int i=0;i<osd.len;i++)

static timestamp_t now;

/*** Options ***/
static char *default_color = "green";
static char *default_outline_color = "black";
static int default_outline_width = 2;
static int default_duration = 1000;
static int default_min_duration = 250;
static int debug_mode = 1;
static int test_mode;


enum long_opt {
  OPT_TEST = 256,
};

static struct msg *current_msg, *first_msg, *last_msg;


static VECTOR(struct osd_creator_abstract) creator=NULL;

static const char short_opts[] = "+c:d:Df:m:o:O:s:";
static const struct option long_opts[] = {
  { "color",		required_argument,	NULL,	'c' },
  { "debug",		no_argument,		NULL,	'D' },
  { "duration",		required_argument,	NULL,	'd' },
  { "font",		required_argument,	NULL,	'f' },
  { "min-duration",	required_argument,	NULL,	'm' },
  { "outline-color",	required_argument,	NULL,	'o' },
  { "outline-width",	required_argument,	NULL,	'O' },
  { "line-spacing",	required_argument,	NULL,	's' },
  { "test",		no_argument,		NULL,	OPT_TEST },	// Undocumented test mode
  { NULL,		0,			NULL,	0   },
};

static void NONRET
usage(void)
{
  fprintf(stderr, "Usage: osdd <options> <modules>\n\n\
Options:\n\
-c, --color=<c>\t\tDefault color (#rgb, #rrggbb or a name from rgb.txt)\n\
-D, --debug\t\tDebugging mode (do not detach from the terminal)\n\
-d, --duration=<ms>\tDefault message duration in milliseconds\n\
-m, --min-duration=<ms>\tDefault minimum message duration in milliseconds\n\
-o, --outline-color=<c>\tDefault outline color\n\
-O, --outline-width=<n>\tDefault outline width (default=2)\n\
\n\
Modules:\n\
Each module id specify by:\n\
\tMODULE_NAME user_name1 user_name2 ... [ args ]\n\
\n\
MODULE_NAME is name from following list\n\
user_name is name of module instance, that could by used by osdc --output=\n\
\t\tonly small letters numbers and _ is allowed\n\
optional arg is really in '[' ']'. Backet should be separate arguments\n\
\n\
");
  for(int i=0;i<VSIZE(creator);i++) if(creator[i].help) creator[i].help(stderr);
  exit(1);
}

static bool 
is_small_ch(char in)
{
	return 'a'<=in && in<='z';
}

static void
parse_opts(int argc, char **argv)
{
  int opt;
  optind = 0;
  while ((opt = getopt_long(argc, argv, short_opts, long_opts, NULL)) >= 0)
  {
	switch (opt)
	  {
	  case 'c':
	default_color = optarg;
	break;
	  case 'd':
	default_duration = atoi(optarg);
	break;
	  case 'D':
	debug_mode = 1;
	break;
	  case 'm':
	default_min_duration = atoi(optarg);
	break;
	  case 'o':
	default_outline_color = optarg;
	break;
	  case 'O':
	default_outline_width = atoi(optarg);
	break;
	  case OPT_TEST:
	test_mode = 1;
	debug_mode = 1;
	break;
	  default:
	fprintf(stderr,"Option %c not exist\n\n",opt);
	usage();
	  }
  }
  int ind = optind;
  while(ind < argc)
  {
	    for(int i=0;i<VSIZE(creator);i++)
	    {
	        if(!strcmp(creator[i].name,argv[ind]))
			{
				char ** name_argv = argv + ++ind;
				int name_argc=0;
				char ** creator_argv = 0;
				int creator_argc=0;
				for(;ind<argc && is_small_ch(argv[ind][0]);ind++) name_argc++;
				if(ind<argc && !strcmp(argv[ind],"["))
				{
					creator_argv = argv + ++ind;
					int open_bracket = 1;
					for(;ind<argc && open_bracket;ind++)
					{
						if(!strcmp(argv[ind],"[")) open_bracket++;
						if(!strcmp(argv[ind],"]")) open_bracket--;
						if(open_bracket) creator_argc++;
					}
					if(open_bracket)
					{
						fprintf(stderr,"Missing ']'\n\n");
						usage();
					}
				}
				char * default_str="default";
				if(name_argc == 0)
				{
					name_argc=1;
					name_argv=&default_str;
				}
				if(creator[i].t == OSD_CEATE_ADD_ONE_OSD)
					osd_set_add(&osd, creator[i].new.add_one_osd(creator_argc,creator_argv,dpy), name_argc, name_argv);
				else
					creator[i].new.add_multiple_osd(&osd,creator_argc,creator_argv,dpy, name_argc, name_argv);
				goto parse_opt_next;
			}
	    }
		fprintf(stderr,"Module %s not exist\n\n",argv[ind]);
		usage();
parse_opt_next:;
  }

}

static char *
osdd_command_get_word(char ** in)
{
	char * a = *in;
	while(**in && **in!=':') ++*in;
	if(**in)
	{
		**in=0;
		++*in;
	}
	return a;
}

static void
osdd_command(VECTOR(char *) cmd_v)
{
	for(int i=0; i<VSIZE(cmd_v); i++)
	{
		char * cmd = cmd_v[i];
		char * a = osdd_command_get_word(&cmd);
		if(!strcmp(a,"enable"))
		{
			osd_set_disable_outputs(&osd, cmd, 0);
		}
		else if(!strcmp(a,"disable"))
		{
			osd_set_disable_outputs(&osd, cmd, 1);
		}
		else if(!strcmp(a,"skip"))
		{
			if(current_msg)
			{
				current_msg->min_duration = current_msg->max_duration = 1;
			}
		}
	}
}

/*** Displaying of messages ***/


static void
display_msg(struct msg *msg)
{
  DBG("## Displaying message\n");
  msg->start_light = now;

  osd_set_show(&osd, msg);
}

static struct msg *
parse_msg(unsigned char *buf, int len)
{
  char *fg_color = default_color;
  char *outline_color = default_outline_color;
  int outline_width = default_outline_width;

  char *line = xmalloc(sizeof(char) * (len+1));
  memcpy(line, buf, len);
  line[len] = 0;

  struct msg *msg = osd_set_new_msg(&osd);
  msg->min_duration = default_min_duration;
  msg->max_duration = default_duration;
  msg->alocated_data=line;

  bool is_output_set=0;
  while (line && *line)
    {
      // The parser is destructive, but it does not do any harm, since we display each message only once.
      char *nl = strchr(line, '\n');
      if(nl) *nl++ = 0;
	  else fprintf(stderr,"MISSING END NL!\n");

      char *key;
      char *val = strchr(line, ':');
      if (val)
	{
	  key = line;
	  *val++ = 0;
	}
      else
	{
	  key = "";
	  val = line;
	}
      DBG("\t%s:%s\n", key, val);

      struct osd_line *l=0;
      if (!key[0])
	{
	  l = msg_add_line(msg, OSD_TYPE_TEXT);
	  l->u.text = val;
	}
      else if (!strcmp(key, "percentage") || !strcmp(key, "percent"))
	{
	  l = msg_add_line(msg, OSD_TYPE_PERCENTAGE);
	  l->u.percent = atoi(val);
	}
      else if (!strcmp(key, "slider"))
	{
	  l = msg_add_line(msg, OSD_TYPE_SLIDER);
	  l->u.percent = atoi(val);
	}
      else if (!strcmp(key, "duration"))
	msg->max_duration = atoi(val);
      else if (!strcmp(key, "min-duration"))
	msg->min_duration = atoi(val);
      else if (!strcmp(key, "color"))
	fg_color = val;
      else if (!strcmp(key, "outline-color"))
	outline_color = val;
      else if (!strcmp(key, "outline-width"))
	outline_width = atoi(val);
      else if (!strcmp(key, "output")||!strcmp(key, "to"))
	  {
		is_output_set=1;
		msg_active_outputs(&osd,msg,val,1);
	  }
      else if (!strcmp(key, "output-no"))
	  {
	    if(!is_output_set) 
			msg_active_outputs(&osd,msg,"default",1);
		is_output_set=1;
		msg_active_outputs(&osd,msg,val,0);
	  }
      else if (!strcmp(key, "cmd-after"))
		  VPB(msg->osdd_command_after, val)
      else if (!strcmp(key, "cmd-now"))
		  VPB(msg->osdd_command_now, val)
      else if (!strcmp(key, "do-not-que"))
		  msg->do_not_que = 1;
      if (l)
	{
	  l->fg_color = fg_color;
	  l->outline_color = outline_color;
	  l->outline_width = outline_width;
	}

      line = nl;
    }

  if(!is_output_set) 
    msg_active_outputs(&osd,msg,"default",1);
  if(!VSIZE(msg->lines))
  {
	  struct osd_line *l = msg_add_line(msg,OSD_TYPE_TEXT);
	  l->u.text = "";
	  l->fg_color = fg_color;
	  l->outline_color = outline_color;
	  l->outline_width = outline_width;
  }
  if (msg->min_duration > msg->max_duration)
    msg->min_duration = msg->max_duration;
  return msg;
}


static void
hide_msg(struct msg *msg)
{
  DBG("## Hiding message\n");
  osd_set_clear(&osd, msg);
  osdd_command(msg->osdd_command_after);
  msg_free(msg);
}

/*** The message queue ***/


static void
enqueue_msg(unsigned char *buf, int len)
{
  DBG("Received: [%.*s]\n", len, buf);
  if (!len || buf[len-1] != '\n')
    return;

  struct msg *msg = parse_msg(buf,len);
  osdd_command(msg->osdd_command_now);
  osd_set_arrive(&osd, msg);
  if(msg->do_not_que)
  {
	  msg_free(msg);
	  return;
  }
  if (first_msg)
    last_msg->next = msg;
  else
    first_msg = msg;
  last_msg = msg;
  msg->next = NULL;
}

static void
parse_input(unsigned char *buf, int len)
{
#ifdef DEBUG_INPUT
  fprintf(stderr,"PARSE_INPUT\n|");
  fwrite(buf,len,1,stderr);
  fprintf(stderr,"|\n");
#endif
  /* The property might contain several messages concatenated. Split them. */
  while (len > 0)
    {
      if (buf[0] == '\n')
	{
	  buf++, len--;
	  continue;
	}
      int i = 0;
      while (i < len && (buf[i] != '\n' || (i && buf[i-1] != '\n')))
	i++;
      enqueue_msg(buf, i);
      buf += i, len -= i;
    }
}

static void
do_test(void)
{
  unsigned char buf[4096];
  int len = 0;
  int c;

  while ((c = read(0, buf + len, 4096 - len)) > 0)
    len += c;
  if (len)
    enqueue_msg(buf, len);
}

/*** Main loop ***/

static
void sigchld(int a)
{
	(void)a;
	int s;
	wait(&s);
}

int
main(int argc, char **argv)
{
  VPB(creator,display_creator_new());
  VPB(creator,display_by_outputs_creator_new());
  VPB(creator,file_creator_new());
  VPB(creator,mqtt_creator_new());
  VPB(creator,exec_creator_new());
  VPB(creator,system_creator_new());

  signal(SIGCHLD, sigchld);

  setlocale(LC_CTYPE, "");
  XInitThreads();

  dpy = XOpenDisplay(NULL);
  if (!dpy)
    die("Cannot open display");
  Window win = DefaultRootWindow(dpy);

  Atom pty = XInternAtom(dpy, "OSD_QUEUE", False);
  if (!pty)
    die("Cannot intern OSD_QUEUE atom");

  if (test_mode)
    {
      do_test();
      pty = 0;
    }
  else
    {
      XSelectInput(dpy, win, PropertyChangeMask);
      XDeleteProperty(dpy, win, pty);
      XFlush(dpy);
    }

  osd = osd_set_new();
  parse_opts(argc, argv);

  if (!debug_mode)
    {
      pid_t pid = fork();
      if (pid < 0)
	die("Cannot fork: %m");
      if (pid > 0)
        return 0;
      setsid();
    }


  struct pollfd pfd = {
    .fd = ConnectionNumber(dpy),
    .events = POLLIN,
  };

  for (;;)
    {
      now = get_current_time();

      timestamp_t wait_until = now - 1;
      if (!current_msg && first_msg)
	{
	  current_msg = first_msg;
	  first_msg = first_msg->next;
	  display_msg(current_msg);
	}
      if (current_msg)
	{
	  if (first_msg)
	    wait_until = current_msg->min_duration + current_msg->start_light;
	  else
	    wait_until = current_msg->max_duration + current_msg->start_light;
	  if (wait_until <= now)
	    {
	      hide_msg(current_msg);
	      current_msg = NULL;
	      continue;
	    }
	}
      if (test_mode && !current_msg)
	break;

      DBG("... waiting for %d ms\n", (int)(wait_until - now));
      poll(&pfd, 1, wait_until - now);
      if (pfd.revents & POLLIN)
	{
	  while (XPending(dpy))
	    {
	      XEvent ev;
	      XNextEvent(dpy, &ev);
	      if (osd_set_handle_event(&osd, &ev))
		continue;
	      if (ev.type != PropertyNotify)
		continue;
	      XPropertyEvent *p = &ev.xproperty;
	      if (p->window == win && p->atom == pty)
		{
		  Atom pty_type;
		  int pty_fmt;
		  unsigned long pty_items, pty_remains;
		  unsigned char *pty_buf = NULL;
		  XGetWindowProperty(dpy, win, pty, 0, 40960, True, XA_STRING, &pty_type, &pty_fmt, &pty_items, &pty_remains, &pty_buf);
		  if (pty_type == XA_STRING && pty_fmt == 8 && pty_items)
		    parse_input(pty_buf, pty_items);
		  if (pty_buf)
		    XFree(pty_buf);
		}
	    }
	}
    }
}
